# Copyright (C) 2001-2016 Quantum ESPRESSO Foundation

AC_DEFUN([X_AC_QE_CUDA], [

AC_ARG_WITH(cuda_dir,                                                                                                                     [AS_HELP_STRING([--with-cuda-dir=<path>],                                                                                               [specify CUDA installation directory (default is /usr/local/cuda/ - MANDATORY)])],
    [cuda_path="$withval"],                                                                                                                [cuda_path="/usr/local/cuda/"])
AC_ARG_WITH(gpu_arch,                                                                                                                     [AS_HELP_STRING([--with-gpu-arch=<arch>],                                                                                             [(sm_13|sm_20|sm_21|sm_30|sm_35|sm_37) Specify the GPU target architecture (default: all)])],
  [cuda_arch="$withval"],                                                                                                                [cuda_arch=all])


AC_ARG_ENABLE(cuda,
   [AS_HELP_STRING([--enable-cuda],
       [enable CUDA (default: no)])],
   [if  test "$enableval" = "yes" ; then
      use_cuda=1
   else
      use_cuda=0
   fi],
   [use_cuda=0])

AC_ARG_ENABLE(debug_cuda,
   [AS_HELP_STRING([--enable-debug-cuda],
       [enable CUDA debug (default: no)])],
   [if  test "$enableval" = "yes" ; then
      use_debugcuda=1
   else
      use_debugcuda=0
   fi],
   [use_debugcuda=0])

AC_ARG_WITH(magma,
   [AS_HELP_STRING([--with-magma],
       [(yes|no|<path>) Use MAGMA. Self-compile or a <path> can be specified (default: no)])],
   [if  test "$withval" = "yes" ; then
      with_magma=1
   elif  test "$withval" = "no" ; then
      with_magma=0
   else
      with_magma=2
      with_magma_path="$withval"
   fi],
   [with_magma=0])

AC_ARG_WITH(phigemm,
   [AS_HELP_STRING([--with-phigemm],
       [(yes|no|<path>|profile) Use PHIGEMM. Self-compile or a <path> can be specified (default: yes)])],
   [if  test "$withval" = "yes" ; then
      with_phigemm=1
      phigemm_profiling=0
   elif  test "$withval" = "no" ; then
      with_phigemm=0
      phigemm_profiling=0
   elif  test "$withval" = "profile" ; then
      with_phigemm=1
      phigemm_profiling=1
   else
      with_phigemm=2
      with_phigemm_path="$withval"
   fi],
   [with_phigemm=1])

AC_ARG_WITH(pinned_mem,
   [AS_HELP_STRING([--with-pinned-mem],
       [(yes|no) Use pinned memory for accelerated BLAS (default: no )])],
   [if  test "$enableval" = "yes" ; then
      with_pinnedmem=1
   else
      with_pinnedmem=0
   fi],
   [with_pinnedmem=0])

nvcc=$NVCC      # NVIDIA CUDA compiler
nvccflags=$NVCCFLAGS      # Flags for NVIDIA CUDA compiler

have_cuda=0
have_phigemm=0

# CUDA
if test "$use_cuda" -eq 1
then

  # Checking if something weird has been specified ...
  case "$cuda_arch" in
  all | sm_13 | sm_20 | sm_21 | sm_30 | sm_35) ;;
  *)
      AC_MSG_ERROR([ GPU arch must be one of: sm_13, sm_20, sm_21, sm_30, or sm_35 ])
  ;;
  esac

    have_valid_cuda_compiler=1
    AC_ARG_VAR([NVCC], [nvcc compiler to use])

    AC_PATH_PROG([NVCC], nvcc)
    if test -n "$nvcc"; then
        have_valid_cuda_compiler=0
        AC_MSG_WARN([CUDA has not been detected])
    fi

    #AC_CHECK_LIB([cudart], [cuInit], AC_MSG_RESULT([FOUND cuda.so]), AC_MSG_FAILURE([NOT FOUND cuda.so]), [])

  if test "$have_valid_cuda_compiler" -eq 1
  then
      # DISABLED (if nvcc is present, cuda.h has to be present. On CRAY systems AC_CHECK_HEADER[cuda] does not work)
      # AC_CHECK_HEADER([cuda],[], AC_MSG_FAILURE([cuda.h NOT FOUND]), [])

      nvcc=$NVCC

      cuda_path=`dirname $NVCC`
      cuda_path=`dirname $cuda_path`

      cuda_libs="-L$cuda_path/lib64 -lcublas  -lcufft -lcudart"

      try_dflags="$try_dflags -D__CUDA"
        have_cuda=1
    fi

  if test "$have_cuda" -eq 1
  then

      # Empty inits... really necessary?
      magma_include=""
      magma_libs=""
      phigemm_include=""
      phigemm_libs=""
      magma_switch=0
      phigemm_switch=0
      phigemm_symbols=0
      cblas_switch=0

      if test "$use_debug" -eq 1; then
         nvccflags="-O0 -g"
      else
         nvccflags="-O3"
      fi

      # GPU architecture -- if not specified compiles for everything, (not ideal)!
    magma_gpu_target="Tesla Fermi kepler"
    case "$cuda_arch" in
         sm_37 )
      nvccflags="$nvccflags -gencode arch=compute_37,code=sm_37"
      magma_gpu_target="Kepler"
            ;;
        sm_35 )
      nvccflags="$nvccflags -gencode arch=compute_35,code=sm_35"
      magma_gpu_target="Kepler"
            ;;
        sm_30 )
      nvccflags="$nvccflags -gencode arch=compute_30,code=sm_30"
      magma_gpu_target="Kepler"
            ;;
        sm_21 )
      nvccflags="$nvccflag -gencode arch=compute_20,code=sm_21"
            magma_gpu_target="Fermi"
            ;;
        sm_20 )
      nvccflags="$nvccflag -gencode arch=compute_20,code=sm_20"
            magma_gpu_target="Fermi"
            ;;
        sm_13 )
        nvccflags="$nvccflags -gencode arch=compute_13,code=sm_13"
            magma_gpu_target="Tesla"
            ;;
        all )
      nvccflags="$nvccflags -gencode arch=compute_13,code=sm_13 -gencode arch=compute_20,code=sm_20 -gencode arch=compute_20,code=sm_21 -gencode arch=compute_30,code=sm_30 -gencode arch=compute_35,code=sm_35"
            magma_gpu_target="Kepler Fermi Tesla"
            ;;
    esac

      cuda_include="-I$cuda_path/include"

      if test "$use_debugcuda" -eq 1 ; then
          try_dflags="$try_dflags -D__CUDA_DEBUG"
      fi

      # phiGEMM - INTERNAL, self-compiled  (with or without profiling)
      if test "$with_phigemm" -eq 1  ; then
          phigemm_switch=1
          phigemm_symbols=1
          try_dflags="$try_dflags -D__PHIGEMM"
          phigemm_opts="-D__PHIGEMM_WEAK_INTERFACES -D__PHIGEMM_ENABLE_SPECIALK"
          phigemm_include="-I$topdir/phiGEMM/include"
          phigemm_libs="$topdir/phiGEMM/lib/libphigemm.a "

    phigemm_cc=$cc
    phigemm_fc=$f90

    if test "$use_debug" -eq 1; then
      phigemm_ccflags="-g -O0 $try_fflags_openmp -fPIC"
            phigemm_ldflags="-g -O0 $try_fflags_openmp -fPIC"
          else
      phigemm_ccflags="-O3 $try_fflags_openmp -fPIC"
            phigemm_ldflags="-O3 $try_fflags_openmp -fPIC"
          fi

          # NVCC HOST-side compiler options
          case "$arch:$f90_version" in
    *:ifort* )
            phigemm_ldsharedflag="--shared"
                phigemm_fflags="-module ../include/"
                phigemm_ldflags="$phigemm_ldflags -zmuldefs"
          ;;
    *:pgf90 | cray*:pgf )
            phigemm_ldsharedflag="-shared"
            phigemm_fflags="-module ../include/"
            phigemm_ldflags="$phigemm_ldflags -Wl,-z,muldefs"
            ;;
    crayxt*:cray* )
      phigemm_ldsharedflag="-shared"
      phigemm_fflags="-em -J../include/"
      phigemm_ccflags="$phigemm_ccflags -shared"
            phigemm_ldflags="$phigemm_ldflags -Wl,-z,muldefs"
      ;;
    *:gfortran )
            phigemm_ldsharedflag="-shared"
            phigemm_fflags="-J../include/"
            phigemm_ldflags="$phigemm_ldflags -z muldefs"
            ;;
    esac

          if test "$use_openmp" -eq 1 ; then
            phigemm_nvccflags="-ccbin gcc -O3 --compiler-options '-c -fPIC -fopenmp'"
          else
                  phigemm_nvccflags="-ccbin gcc -O3 --compiler-options '-c -fPIC'"
    fi



    if test "$with_pinnedmem" -eq 1  ; then
            phigemm_opts="$phigemm_opts -D__PHIGEMM_PINNED"
      try_dflags="$try_dflags -D__CUDA_MEM_PINNED"
          fi

          if test "$phigemm_profiling" -eq 1; then
                  phigemm_opts="$phigemm_opts -D__PHIGEMM_PROFILE"
                  try_dflags="$try_dflags -D__PHIGEMM_PROFILE"
          fi
      fi

      # phiGEMM - EXTERNAL, assumed well-compiled (no profiling)
      if test "$with_phigemm" -eq 2  ; then
          phigemm_switch=0
          phigemm_symbols=1
          try_dflags="$try_dflags -D__PHIGEMM"
          phigemm_include="-I$with_phigemm_path/include"
          phigemm_libs="$with_phigemm_path/lib/libphigemm.a "
      fi

      # By default show all CUDA timing
      # try_dflags="$try_dflags -D__CUDA_QE_TIMING"

      # MAGMA - INTERNAL, self-compiled
      if test "$with_magma" -eq 1 ; then
          magma_switch=1
          magma_include="-I$topdir/qe-magma/include -I$topdir/quark/include"
          magma_libs="$topdir/qe-magma/lib/libmagma.a "

    try_dflags="$try_dflags -D__MAGMA -DHAVE_CUBLAS"

    magma_openmp=""
                if test "$use_openmp" -eq 1 ; then
                        magma_openmp="-fopenmp"
                fi

    magma_dflags="-DMAGMA_SETAFFINITY"

    if test "$have_mkl" -eq 1
    then
      magma_dflags="$magma_dflags -DMAGMA_WITH_MKL"
    fi

   # CBLAS  (usually MKL and AMCL have CBLAS interfaces natively)
          # This does not work on CRAY
        have_cblas=0
        if test "$have_mkl" -eq 0 && test "$have_acml" -eq 0 && test "$openblas_switch" -eq 0
        then
          try_libdirs=""
              try_libdirs="$libdirs $ld_library_path $try_libdirs"
            for dir in $try_libdirs
            do
                unset ac_cv_search_cblas_dscal # clear cached value

                if test "$dir" = "none"
                then
                        try_loption=
                else
                        try_loption="-L$dir"
                fi
                  FFLAGS="$test_fflags"
                  LDFLAGS="$test_ldflags $try_loption"
                  LIBS=""

                LIBS=""
                echo $ECHO_N "in $dir: " $ECHO_C
                AC_SEARCH_LIBS(cblas_dscal, cblas,
                     have_cblas=1 magma_libs="$magma_libs $LIBS")

                if test "$have_cblas" -eq 1
                then
                    break
                fi
            done
            if test "$have_cblas" -eq 0
            then
              magma_libs="$magma_libs $topdir/CBLAS/lib/cblas.a"
              cblas_switch=1
              have_cblas=1
              case "$arch:$f90_version" in
          *:gfortran )
                  extlib_flags="$extlib_flags -fPIC"
          ;;
              esac
              fi
          fi
          # This if CRAY and LibSCi are used, LibSCi does not contains CBLAS
          # interfaces. I hope to find a way to detect this case automatically
          # (NdFilippo)
          if test "$use_internal_cblas" -eq 1 ; then
            magma_libs="$magma_libs $topdir/CBLAS/lib/cblas.a"
            cblas_switch=1
            have_cblas=1
          fi
      fi


      # MAGMA - EXTERNAL, assumed well-compiled and all dependency linked (CBLAS)
      if test "$with_magma" -eq 2 ; then
          magma_switch=0
          magma_include="-I$with_magma_path/include -I$with_magma_path/quark/include"
          magma_libs="$with_magma_path/lib/libmagma.a "
          try_dflags="$try_dflags -D__MAGMA -DHAVE_CUBLAS"
      fi

      try_iflags="$try_iflags $magma_include $phigemm_include $cuda_include"

      lapack_libs="$magma_libs $lapack_libs"
      blas_libs="$phigemm_libs $blas_libs"

      ld_libs="$cuda_libs $ld_libs"

      echo setting CUDA_LIBS... $cuda_libs
    fi
fi


# CUDA
AC_SUBST(nvcc)
AC_SUBST(nvccflags)

AC_SUBST(cuda_path)
AC_SUBST(cuda_include)
AC_SUBST(cuda_libs)

AC_SUBST(phigemm_cc)
AC_SUBST(phigemm_fc)
AC_SUBST(phigemm_opts)
AC_SUBST(phigemm_include)
AC_SUBST(phigemm_libs)
AC_SUBST(phigemm_symbols)
AC_SUBST(phigemm_switch)
AC_SUBST(phigemm_ccflags)
AC_SUBST(phigemm_fflags)
AC_SUBST(phigemm_nvccflags)
AC_SUBST(phigemm_ldflags)
AC_SUBST(phigemm_ldsharedflag)

AC_SUBST(cblas_switch)

AC_SUBST(magma_switch)
AC_SUBST(magma_gpu_target)
AC_SUBST(magma_openmp)
AC_SUBST(magma_dflags)

# Third-part CUDA-enabled libraries
AC_CONFIG_FILES(install/make_phiGEMM.inc)
AC_CONFIG_FILES(install/make_magma.inc)
AC_CONFIG_FILES(install/make_cblas.inc)

])
